@@ -12,8 +12,14 @@ from docker.errors import ContainerError
 import paramiko
 from scp import SCPClient
 
-from build_magic.exc import HostWorkingDirectoryNotFound
-from build_magic.reference import BindDirectory, HostWorkingDirectory
+from build_magic.exc import HostWorkingDirectoryNotFound, OSEnvironmentMismatch
+from build_magic.reference import BindDirectory, HostWorkingDirectory, OSVersionCommands
+
+
+HOST_WD = 'hostwd'
+BIND_DIR = 'bind'
+ENVS = 'envs'
+CWD = '.'
 
 
 class Status:
@@ -142,7 +148,16 @@ class CommandRunner:
         - teardown()
     """
 
-    def __init__(self, environment, working_dir='', copy_dir='', timeout=30, artifacts=None, parameters=None):
+    def __init__(
+            self,
+            environment,
+            working_dir='',
+            copy_dir='',
+            timeout=30,
+            artifacts=None,
+            parameters=None,
+            envs=None,
+    ):
         """Instantiate a new CommandRunner object."""
         self.environment = environment
         self.working_directory = working_dir
@@ -150,8 +165,13 @@ class CommandRunner:
         self.timeout = timeout
         self.artifacts = [] if not artifacts else artifacts
         self.parameters = {} if not parameters else parameters
+        self.envs = {} if not envs else envs
 
-        # TODO: Need to validate the types for artifacts and parameters.
+    def _merge_envs(self):
+        """Merge environment variables with self.envs."""
+        envs = os.environ
+        envs.update(self.envs)
+        self.envs = envs
 
     @staticmethod
     def cd(directory):
@@ -181,9 +201,10 @@ class CommandRunner:
         """
         if not self.artifacts:
             return False
+
         if src and dst:
             for artifact in self.artifacts:
-                if src == '.':
+                if src == CWD:
                     src = Path.cwd().resolve()
                 source = Path(src) / artifact
                 try:
@@ -238,9 +259,47 @@ class CommandRunner:
 class Local(CommandRunner):
     """Manages macros executed on the local host machine."""
 
-    def __init__(self, environment='', working_dir='', copy_dir='', timeout=30, artifacts=None, parameters=None):
+    def __init__(
+            self,
+            environment='',
+            working_dir='',
+            copy_dir='',
+            timeout=30,
+            artifacts=None,
+            parameters=None,
+            envs=None,
+    ):
         """Instantiates a new Local command runner object."""
-        super().__init__(environment, working_dir, copy_dir, timeout, artifacts, parameters)
+        super().__init__(environment, working_dir, copy_dir, timeout, artifacts, parameters, envs)
+
+    def os_matches_environment(self):
+        """Tests if self.environment matches the locally running operating system or distro.
+
+        :return: True if the environment matches the OS, else False.
+        """
+        if self.environment.lower() in ('macos', 'darwin'):
+            cmd = OSVersionCommands.DARWIN.value
+        elif self.environment.lower() in ('win', 'windows'):
+            cmd = OSVersionCommands.WINDOWS.value
+        else:
+            cmd = OSVersionCommands.LINUX.value
+
+        result = subprocess.run(
+            cmd,
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+            shell=True,
+        )
+
+        if result.returncode > 0:
+            return False
+
+        if self.environment.lower() in ('macos', 'darwin'):
+            return True if b'Mac OS X' or b'MacOS' in result.stdout else False
+        elif self.environment.lower() in ('win', 'windows'):
+            return True if b'Microsoft Windows' in result.stdout else False
+        else:
+            return True if bytes(self.environment.lower(), encoding='utf-8') in result.stdout else False
 
     def prepare(self):
         """Changes to the specified working directory and copies artifacts if necessary.
@@ -250,8 +309,14 @@ class Local(CommandRunner):
         """
         if self.copy_from_directory:
             self.copy(self.copy_from_directory, self.working_directory)
+
         if self.working_directory:
             self.cd(self.working_directory)
+
+        if self.environment:
+            if self.os_matches_environment() is False:
+                raise OSEnvironmentMismatch
+
         return True
 
     def execute(self, macro):
@@ -262,13 +327,15 @@ class Local(CommandRunner):
         :return: A Status object reflecting the results of the macro.
         """
         command = macro.as_string()
+
         result = subprocess.run(
             command,
             stdout=subprocess.PIPE,
             stderr=subprocess.PIPE,
-            # timeout=self.timeout,
+            env=self._merge_envs(),
             shell=True,
         )
+
         return Status(result.stdout, result.stderr, result.returncode)
 
 
@@ -286,12 +353,13 @@ class Remote(CommandRunner):
             timeout=30,
             artifacts=None,
             parameters=None,
+            envs=None,
     ):
         """Instantiates a new Remote command runner object."""
-        if working_dir == '.':
+        if working_dir == CWD:
             working_dir = ''
 
-        super().__init__(environment, working_dir, copy_dir, timeout, artifacts, parameters)
+        super().__init__(environment, working_dir, copy_dir, timeout, artifacts, parameters, envs)
 
         self.user = None
         self.host = None
@@ -350,10 +418,12 @@ class Remote(CommandRunner):
         ssh = paramiko.SSHClient()
         ssh.load_system_host_keys()
         conn_args = dict(hostname=self.host, pkey=self.key)
+
         if self.port:
             conn_args.update(dict(port=self.port))
         if self.user:
             conn_args.update(dict(username=self.user))
+
         try:
             ssh.connect(**conn_args)
         except socket.gaierror:
@@ -371,6 +441,7 @@ class Remote(CommandRunner):
         # Add the absolute path to each artifact.
         src = Path(src)
         files = []
+
         for artifact in self.artifacts:
             files.append(str(src / artifact))
 
@@ -416,8 +487,14 @@ class Remote(CommandRunner):
         command = macro.as_string()
 
         ssh = self.connect()
+
         try:
-            stdin_, stdout_, stderr_ = ssh.exec_command(command, timeout=self.timeout)
+            stdin_, stdout_, stderr_ = ssh.exec_command(
+                command,
+                timeout=self.timeout,
+                get_pty=True,
+                environment=self.envs,
+            )
             stdout = stdout_.readlines()
             stderr = stderr_.readlines()
             exit_code = stdout_.channel.recv_exit_status()
@@ -427,35 +504,120 @@ class Remote(CommandRunner):
             )
         finally:
             ssh.close()
+
         return Status(stdout=''.join(stdout), stderr=''.join(stderr), exit_code=exit_code)
 
 
 class Vagrant(CommandRunner):
     """Manages macros executed executed in a guest virtual machine managed by Vagrant."""
 
+    vagrantfile_name = 'Vagrantfile'
+    alt_vagrantfile_name = 'Vagrantfile_build_magic'
+    vagrant_cwd_env = 'VAGRANT_CWD'
+    vagrantfile_name_env = 'VAGRANT_VAGRANTFILE'
+    default_home_dir = '/home/vagrant'
+    default_bind_dir = '/vagrant'
+
     def __init__(
             self,
-            environment='.',
-            working_dir='/home/vagrant',
+            environment=CWD,
+            working_dir='',
             copy_dir='',
             timeout=30,
             artifacts=None,
             parameters=None,
+            envs=None,
     ):
         """Instantiates a new Vagrant command runner object."""
-        super().__init__(environment, working_dir, copy_dir, timeout, artifacts, parameters)
+        if not working_dir:
+            working_dir = self.default_home_dir
+
+        super().__init__(environment, working_dir, copy_dir, timeout, artifacts, parameters, envs)
+
         self._vm = None
-        if self.environment == 'Vagrantfile':
-            self.environment = '.'
-        if self.environment != '.':
-            if 'Vagrantfile' in self.environment:
-                self.environment = self.environment.split('Vagrantfile')[0]
-            os.environ.pop('VAGRANT_CWD', '')
-            os.environ['VAGRANT_CWD'] = str(Path(self.environment).resolve())
-        self.host_wd = self.parameters.get('hostwd', HostWorkingDirectory(self.environment)).value
+        self._vagrantfile_config = {
+            BIND_DIR: False,
+            ENVS: False,
+        }
+
+        if self.environment == self.vagrantfile_name:
+            self.environment = CWD
+
+        if self.environment != CWD:
+            if self.vagrantfile_name in self.environment:
+                self.environment = self.environment.split(self.vagrantfile_name)[0]
+            os.environ.pop(self.vagrant_cwd_env, '')
+            os.environ[self.vagrant_cwd_env] = str(Path(self.environment).resolve())
+
+        self.bind_path = self.parameters.get(BIND_DIR, BindDirectory(self.default_bind_dir)).value
+        if self.bind_path != self.default_bind_dir:
+            self._vagrantfile_config[BIND_DIR] = True
+
+        if self.envs:
+            self._vagrantfile_config[ENVS] = True
+
+        self.host_wd = self.parameters.get(HOST_WD, HostWorkingDirectory(self.environment)).value
         if not Path(self.host_wd).exists():
             raise HostWorkingDirectoryNotFound
-        self.bind_path = self.parameters.get('bind', BindDirectory('/vagrant')).value
+
+        # Create a new Vagrantfile if needed.
+        if any(self._vagrantfile_config.values()):
+            config = self.build_config()
+            self.create_vagrantfile_config(config)
+
+    def build_config(self):
+        """Assemble the Vagrantfile config based on the command runner settings.
+
+        :rtype: str
+        :return: The Vagrantfile config.
+        """
+        wd = self.default_home_dir if self.working_directory == CWD else self.working_directory
+
+        bind_dir = ''
+        if self._vagrantfile_config.get(BIND_DIR, False):
+            bind_dir = f'  config.vm.synced_folder "{CWD}", "{self.bind_path}"'
+
+        envs = []
+        exports = ''
+        if self.envs:
+            for env, value in self.envs.items():
+                envs.append(f'echo "export {env}={value}" >> {wd}/.profile')
+            exports = '\n'.join(envs)
+
+        provision = ''
+        if exports:
+            provision = f"""  config.vm.provision "build-magic", type: "shell" do |s|
+    s.inline = <<-SCRIPT
+{exports}
+SCRIPT
+  end"""
+
+        config = f"""# Config added by build-magic
+Vagrant.configure("2") do |config|
+{bind_dir}
+{provision}
+end"""
+        return config
+
+    def create_vagrantfile_config(self, config):
+        """Concatenate a new Vagrantfile config to an existing Vagrantfile.
+
+        The steps are:
+
+            1. Read the Vagrantfile from self.environment.
+            2. Append config to the end of the Vagrantfile.
+            3. Write the new Vagrantfile to the same directory as the existing Vagrantfile.
+            4. Set the VAGRANT_VAGRANTFILE environment variable to point to the new Vagrantfile.
+
+        :param config: A Vagrant.config section to add to an existing Vagrantfile.
+        :return: None
+        """
+        vagrantfile_path = Path(self.environment).resolve()
+        vagrantfile = vagrantfile_path.joinpath(self.vagrantfile_name).read_text()
+        new_vagrantfile = vagrantfile + config
+        vagrantfile_path.joinpath(self.alt_vagrantfile_name).write_text(new_vagrantfile)
+        os.environ.pop(self.vagrantfile_name_env, '')
+        os.environ[self.vagrantfile_name_env] = self.alt_vagrantfile_name
 
     def prepare(self):
         """Handles copying artifacts to the working directory if necessary."""
@@ -464,6 +626,7 @@ class Vagrant(CommandRunner):
                 self._vm.ssh(command=f'sudo mkdir {self.working_directory}')
             except subprocess.CalledProcessError:
                 return False
+
         if self.copy_from_directory and len(self.artifacts) > 0:
             self.copy(Path(self.copy_from_directory).resolve(), Path(self.host_wd).resolve())
             try:
@@ -479,9 +642,11 @@ class Vagrant(CommandRunner):
         :param Macro macro: The Macro object to execute.
         :return: The Status of the executed Macro.
         """
-        if self.working_directory != '/home/vagrant':
+        if self.working_directory != self.default_home_dir:
             macro.prefix = f'cd {self.working_directory};'
+
         cmd = macro.as_string()
+
         try:
             out = self._vm.ssh(command=cmd)
         except subprocess.CalledProcessError as err:
@@ -492,27 +657,33 @@ class Vagrant(CommandRunner):
 class Docker(CommandRunner):
     """Manages macros executed in a Docker container."""
 
+    default_working_dir = '/build_magic'
+
     def __init__(
             self,
             environment='alpine',
-            working_dir='/build_magic',
+            working_dir='',
             copy_dir='',
             timeout=30,
             artifacts=None,
-            parameters=None
+            parameters=None,
+            envs=None,
     ):
         """Instantiates a new Docker command runner object."""
-        super().__init__(environment, working_dir, copy_dir, timeout, artifacts, parameters)
-        if self.working_directory == '.':
-            self.working_directory = '/build_magic'
-        self.host_wd = self.parameters.get('hostwd', HostWorkingDirectory('.')).value
+        if not working_dir or working_dir == CWD:
+            working_dir = self.default_working_dir
+
+        super().__init__(environment, working_dir, copy_dir, timeout, artifacts, parameters, envs)
+
+        self.host_wd = self.parameters.get(HOST_WD, HostWorkingDirectory(CWD)).value
         if not Path(self.host_wd).exists():
             raise HostWorkingDirectoryNotFound
-        self.bind_path = self.parameters.get('bind', BindDirectory()).value
+
+        self.bind_path = self.parameters.get(BIND_DIR, BindDirectory()).value
         self.binding = Mount(
             target=self.bind_path,
             source=str(Path(self.host_wd).resolve()),
-            type='bind',
+            type=BIND_DIR,
         )
         self.client = None
         self.container = None
@@ -551,6 +722,7 @@ class Docker(CommandRunner):
                 stdout=True,
                 stderr=True,
                 tty=True,
+                environment=self.envs,
             )
             if out:
                 out = out.decode('utf-8')
